/*     Loglan82 Compiler&Interpreter
     Copyright (C) 1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#ifndef __GENINT_H
#define __GENINT_H

/**
 * @file
 */

#if GEN
#define word address
#endif


/**
 * Appetites of values
 * \defgroup AppetitesOfValues Appetites of values
 * @{
 */
/** integer,boolean,string,char */
#define APINT           1

/** real */
#define APREAL          sizeof(real)/sizeof(word)

/**
 * To tell apart a virtual address
 * (any reference type) from any other type
 */
#define APVIRT          4

/** actual size of reference type */
#define APREF           2

/** formal type value */
#define APFMTYPE        2

/** formal procedure or function value, = apref + 1 */
#define APFMPROC        3

#define APOPCODE        (sizeof(extopcode) + sizeof(word) - 1) / sizeof(word)

/** @} */

#ifndef __cplusplus
typedef int bool;
#endif

/* In Qt4 these 2 definitions are already made */
#ifndef FALSE
#define FALSE           0
#endif

#ifndef TRUE
#define TRUE            1
#endif


/* Values identifying types and formal parameters: */
/** @{ */

#define FORMFUNC        7       /* parameter "function" */
#define FORMPROC        8       /* parameter "procedure" */
#define FORMTYPE        9       /* parameter "type", also formal type */
#define PARIN           11      /* parameter "input" */
#define PAROUT          12      /* parameter "output" */
#define PARINOUT        16      /* parameter "inout" */
#define PUREPROCESS     13      /* type PROCESS */
#define PURECOROUTINE   14      /* type COROUTINE */
#define PRIMITIVETYPE   15      /* simple type */
#define CLASSTYPE       17      /* class type */
#define FILETYPE        18      /* file type */
/** @} */

/* Values to identify kinds of array elements : */
/** @{ */
#define AINT            -1      /* arrayof integer */
#define AREAL           -3      /* arrayof real */
#define AVIRT           -2      /* arrayof reference */
#define APROCESS        -4      /* process reference implemented as arrayof */
#define FILEOBJECT      -11     /* file object */
#define SKILLED         -99     /* killed object, only for compactification */

/** @} */

#define DUMMY           -1      /* dummy unit or prefix : SL for main block */
#define MAINBLOCK       0       /* main block prototype number */
#define MAXPROT         500     /* limit for the number of prototypes */


/* L-code instruction addressing modes : */
/** @{ */
#define GLOBAL          0       /* M[ N ]                */
#define LOCAL           1       /* M[ c1+N ]             */
#define TEMPLOCAL       2       /* M[ c2+N ]             */
#define REMOTE          3       /* M[ M[ K ]+N ]         */
#define INDIRECT        4       /* M[ M[ N ] ]           */
#define IMMEDIATE       5       /* N                     */
#define CONSTANT        6       /* M[ N ]                */
#define DOTACCESS       7       /* M[ M[ display+K ]+N ] */
#define NOARGUMENT      8       /* ------                */
/** @} */


/* Prototype kind : */
/** @{ */
#define CLASS           0
#define RECORD          1
#define COROUTINE       2
#define PROCESS         3
#define BLOCK           4
#define PREFBLOCK       5
#define FUNCTION        6
#define PROCEDURE       7
#define HANDLER         8
/** @} */

#if GEN
#define LRECORD    RECORD
#define LFUNCTION  FUNCTION
#define LPROCEDURE PROCEDURE
#endif


#if GEN
typedef int protaddr;   /* mainblock..maxprot ; mainblock = 0; */
                             /* INDIRECT ADDRESS OF PROTOTYPE       */
typedef int dprotaddr;  /* PROTADDR + DUMMY                    */
typedef int protkind;
#else
#define protaddr  int
#define dprotaddr int
#define protkind  int
#endif

/* TODO: Conditionally restore it:*/
typedef int word;

/**
 * Prototype:
 * \defgroup Prototype Prototype
 * @{
 */
typedef struct {
	/**
	 * Prototype of syntactic father
	 * or dummy for the main block
	 */
	dprotaddr slprototype;

	/** address of code */
	word codeaddr;
	/** object's total length */
	word appetite;
	/**
	 * distance from beginning of object to
	 * temporary reference variables
	 */
	word span;

	/** beginning and length of table with */
	word reflist;

	/**
	 * offsets of reference values in object;
	 * without SL, DL, CL and temporary values
	 */
	int lthreflist;
	/** beginning and length of table with */
	word parlist;
	/** offsets of parameters */
	int lthparlist;

	/** beginning and length of table with */
	word preflist;
	/** prototype numbers of prefix sequence */
	int lthpreflist;

	/** lastwill code address */
	word lastwill;
	/** list of handlers, 0 if empty */
	word handlerlist;
	/**
	 * address of table for conversion:
	 * virtual number --> prototype
	 */
	word virtlist;

	/**
	 * only for processes:
	 * @{
	 */
	/** lowest procedure prototype number */
	protaddr maskbase;
	/** size of RPC mask for that process */
	int masksize;
	/** @} */

	/**
	 * for procedures and functions
	 * \defgroup PrototypeForFunctions
	 * @{
	 */
	/* virtual number or -1 if not virtual */
	int virtnumber;
	/** @} */

	/**
	 * for all :
	 * \defgroup PrototypeForAll
	 * @{
	 */
	/** kind of unit */
	protkind  kind;
	/** @} */

	/**
	 * only for procedures, functions and processes:
	 * \defgroup PrototypeForProceduresFunctionsProcesses
	 * @{
	 */
	/** address of table with parameters descr. addresses */
	word pfdescr;
	/** @} */

	/**
	 * only for functions:
	 * \defgroup PrototypeForFunctions
	 * @{
	 */
	int nrarray;
	/** @} */
	/* type of result */
	word finaltype;

} protdescr,*pointprdsc;

/** @} */

#endif /* __GENINT_H */
