#include <QtGui/QApplication>
#include <QtGui/QMainWindow>
#include <Qt3Support/q3multilineedit.h>
#include <QtGui/QMenuBar>
// #include <qpopmenu.h>
#include <QtGui/QDialog>
#include <QtGui/QButtonGroup>
#include <QtGui/QLabel>
#include <QtGui/QLineEdit>
#include <Qt3Support/Q3ListBox>
#include <QtGui/QPushButton>
#include <QtGui/QRadioButton>
#include <QtCore/QList>
#include <QtCore/qfile.h>
#include <QtGui/qcombobox.h>
#include <QtGui/QToolTip>
#include <QtGui/QFont>
#include <QtGui/QPixmap>
#include <QtGui/QMessageBox>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <QtCore/QTextStream>
#include <QtCore/qstring.h>
#include <QtGui/QFileDialog>
#include <QtGui/QFontMetrics>
#include <QtGui/QPainter>
#include <QtGui/QScrollBar>


#define TAG_TEXT	0
#define TAG_OTHER	1

#define TAG_LINK	2 
#define TAG_ANCHOR	3 
#define TAG_BOLD_ON	4 
#define TAG_BOLD_OFF	5 
#define TAG_ITALIC_ON	6 
#define TAG_ITALIC_OFF  7 
#define TAG_BREAK	8 
#define TAG_LINE	9 
#define TAG_LINK_END	10 
#define TAG_LIST_ON	11
#define TAG_LIST_OFF	12
#define TAG_LIST_ITEM	13

#define PIX_HEIGHT	2000

char names[14][40] = {
	"text", "other", "link", "anchor", "bold on",
	"bold off", "italic on", "italic off", "break",
	"line", "end link", "list on", "list off", "list item"
};

class Tag {
public:
	Tag() {
		tag_type=0;
		strcpy(tag_label, "");
		strcpy(tag_link, "");
		strcpy(tag_text, "");
	};
	int tag_type;
	char tag_label[255];
	char tag_link[255];
	char tag_text[255]; 
	int x;
	int y;
	int w;
	int h;
};

class HTMLAnalyzer {
public:
	QList<Tag*> tags;
	bool verbatim;

	HTMLAnalyzer();

	bool LoadFile(char*);
	void AnalyzeTag(QString*);
	void PackLinks();
	void DumpList();
	Tag *CheckTag(int, int);
	Tag *FindAnchor(char*);
};

HTMLAnalyzer::HTMLAnalyzer()
{
	tags.clear();
	verbatim = FALSE;
}

bool HTMLAnalyzer::LoadFile(char *fname)
{
	QFile f(fname);
	QString poms, poms1;
	int i;
	Tag *pomt;
	bool not_ended;

	tags.clear();
	if (!f.open(IO_ReadOnly))
		return(FALSE);

	QTextStream fs(&f);

	while (!fs.atEnd()) {
		poms = fs.readLine();
		while (poms.length()>0) {
			i = poms.find('<');
			if (i != -1) {
				if (i > 1) {
					poms1 = poms.left(i);
					pomt = new Tag;
					pomt->tag_type = TAG_TEXT;
					if (!verbatim) {
						poms1 = poms1.simplifyWhiteSpace();
					}
						
					sprintf(pomt->tag_text, " %s", poms1.toStdString().c_str());
					tags.append(pomt);
				}
				poms = poms.right(poms.length() - i);

				i = poms.find('>');
				if (i != -1) {
					poms1 = poms.mid(1, i - 1);
					AnalyzeTag(&poms1);
					poms = poms.right(poms.length() - i -1);
				} else {
					not_ended = TRUE;
					break;
				}
			} else {
				pomt = new Tag;
				pomt->tag_type = TAG_TEXT;
				if (!verbatim) 
					poms=poms.simplifyWhiteSpace();
				
				sprintf(pomt->tag_text, " %s", poms.toStdString().c_str());
				tags.append(pomt);
				break;
			}
		}
		if (verbatim) {
			pomt = new Tag;
			pomt->tag_type = TAG_BREAK;
			tags.append(pomt);
		}
	}
	f.close();
	return TRUE;
}

void HTMLAnalyzer::AnalyzeTag(QString *t)
{
	Tag *pom;
	int i;
	QString poms, poms1;

	*t = t->simplifyWhiteSpace();
	pom = new Tag;

	if ((t->toStdString().c_str()[0] != 'A') && (t->toStdString().c_str()[0] != 'a')) {
		*t = t->upper();
		if (strcmp(t->toStdString().c_str(), "B") ==0 ) {
			pom->tag_type = TAG_BOLD_ON;
		}
		else if (strcmp(t->toStdString().c_str(), "/B") == 0) {
			pom->tag_type = TAG_BOLD_OFF;
		}
		else if (strcmp(t->toStdString().c_str(), "I") == 0) {
			pom->tag_type = TAG_ITALIC_ON;
		}
		else if (strcmp(t->toStdString().c_str(), "/I") == 0) {
			pom->tag_type = TAG_ITALIC_OFF;
		}
		else if (strcmp(t->toStdString().c_str(), "BR") == 0) {
			pom->tag_type = TAG_BREAK;
		}
		else if (strcmp(t->toStdString().c_str(), "HR") == 0) {
			pom->tag_type = TAG_LINE;
		}
		else if (strcmp(t->toStdString().c_str(), "/A") == 0) {
			pom->tag_type = TAG_LINK_END;
		}
		else if (strcmp(t->toStdString().c_str(), "UL") == 0) {
			pom->tag_type = TAG_LIST_ON;
		}
		else if (strcmp(t->toStdString().c_str(), "/UL") == 0) {
			pom->tag_type = TAG_LIST_OFF;
		}
		else if (strcmp(t->toStdString().c_str(), "LI") == 0) {
			pom->tag_type = TAG_LIST_ITEM;
		}
		else if (strcmp(t->toStdString().c_str(), "PRE") == 0) {
			verbatim=TRUE;
		}
		else if (strcmp(t->toStdString().c_str(),"/PRE") == 0) {
			verbatim=FALSE;
		}
	} else {
	 /*
	'a' or 'A'
	links
	*/
		i = t->find('=');
		if (i != -1) {
			poms = t->mid(2, i - 2);
			poms = poms.simplifyWhiteSpace();
			poms = poms.upper();
			poms1 = t->right(t->length() - i - 1);
			poms1 = poms1.simplifyWhiteSpace();

			if (poms == "HREF") {
				pom->tag_type = TAG_LINK;
				strcpy(pom->tag_link,poms1.toStdString().c_str());
			}
			else if (poms == "NAME") {
				pom->tag_type = TAG_ANCHOR;
				if (poms1[0] == '"')
					poms1=poms1.right(poms1.length() - 1);
				if (poms1[poms1.length() - 1] == '"')
					poms1=poms1.left(poms1.length() - 1);
				strcpy(pom->tag_label, poms1.toStdString().c_str());
			}
		}
	}
	tags.append(pom);
}

void HTMLAnalyzer::DumpList()
{
	Tag *pom;
	QListIterator<Tag *> tagsIterator(tags);

	while (tagsIterator.hasNext()) {
		pom = tagsIterator.next();
		fprintf(stderr, "%s:%s,%s,%s\n", names[pom->tag_type],
			pom->tag_text, pom->tag_link, pom->tag_label);
	}
}

void HTMLAnalyzer::PackLinks()
{
	Tag *pom,*pom1;
	char s[255];
	QListIterator<Tag *> tagsIterator(tags);

	while (tagsIterator.hasNext()) {
		pom = tagsIterator.next();
		if ((pom->tag_type == TAG_LINK) ||
			(pom->tag_type == TAG_ANCHOR)) {
			pom1 = tagsIterator.next();
			strcpy(s, "");
			while ((pom1 != NULL) &&
				(pom1->tag_type != TAG_LINK_END)) {
				if (pom1->tag_type == TAG_TEXT)
					strcat(s,pom1->tag_text);
				tags.removeOne(pom1);
				pom1 = tags.current();
			}
			strcpy(pom->tag_text,s); 
			tags.removeOne(pom1);
			pom = tags.current();
		} else {
			pom = tagsIterator.next();
		}
	}
}


Tag *HTMLAnalyzer::CheckTag(int x,int y)
{
	Tag *pom;
	QListIterator<Tag *> tagsIterator(tags);

	while (tagsIterator.hasNext()) {
		pom = tagsIterator.next();
		if (pom->tag_type==TAG_LINK)
			if ((x >= pom->x) && (x <= pom->x + pom->w) &&
				(y >= pom->y) && (y <= pom->y + pom->h)) {
				return pom;
			}
	}
	return NULL;
}

Tag *HTMLAnalyzer::FindAnchor(char *name)
{
	Tag *pom;
	QListIterator<Tag *> tagsIterator(tags);

	while (tagsIterator.hasNext()) {
		pom = tagsIterator.next();
		if ((pom->tag_type == TAG_ANCHOR) &&
			(strcmp(pom->tag_label, name) == 0)) {
				return(pom);
			}
	}
	return pom;
}
/********************************/

class QHTML: public QMainWindow {
	Q_OBJECT
public:
	QScrollBar *vscroll;
	QMenuBar *bar;  
	HTMLAnalyzer *analyzer;
	QPixmap *map;
	int cx,cy,oy,lstep,pstep;
	bool Bold,Italic;
	QFont *normal,*bold,*italic,*bold_italic,*actual_font;
	char homedir[255];

	QHTML(char*);
	void DrawList();
	virtual void paintEvent(QPaintEvent *ev);
	virtual void resizeEvent(QResizeEvent *ev);
	virtual void mousePressEvent(QMouseEvent *ev);
public slots:
	void load();
	void back();
	void vscrolled(int);
	void contents();
	void user_guide();
	void lang_guide();
};


QApplication *app;

QHTML::QHTML(char *d)
{
	char s[255];

	QFont f("Helvetica", 12, QFont::Bold);
	normal = new QFont("Helvetica", 12, QFont::Normal);
	bold = new QFont("Helvetica", 12, QFont::Bold);
	italic = new QFont("Helvetica", 12, QFont::Normal, TRUE);
	bold_italic = new QFont("Helvetica", 12, QFont::Bold, TRUE);
	strcpy(homedir, d);

	actual_font = normal;
	
	setWindowTitle("LOGLAN Help");

	QMenu * contentsMenu = NULL;
	menuBar()->addAction("File", this, SLOT(load()));
	contentsMenu = menuBar()->addMenu("&Contents");
	contentsMenu->addAction("Index", this, SLOT(contents()));
	contentsMenu->addAction("User Guide", this, SLOT(user_guide()));
	contentsMenu->addAction("Language reference", this, SLOT(lang_guide()));

	menuBar()->addAction("Quit", app, SLOT(quit()));

	analyzer = new HTMLAnalyzer;
	map = new QPixmap(500, PIX_HEIGHT);
	map->fill(backgroundColor());
	resize(500, height());
	setFixedSize(width(), height());
	oy = 0;
	lstep = 10;
	pstep = height() - bar->height();
	vscroll = new QScrollBar(0, PIX_HEIGHT, lstep, pstep, 0, QScrollBar::Vertical, this);
	vscroll->setTracking(TRUE);  
	vscroll->setGeometry(width() - 16, bar->height(), 16, height()-bar->height());
	connect(vscroll, SIGNAL(valueChanged(int)), this, SLOT(vscrolled(int)));
	sprintf(s, "%s/index.html", homedir);
	analyzer->LoadFile(s);
	analyzer->PackLinks();
	DrawList();
}

void QHTML::vscrolled(int v)
{
	oy = v;
	repaint();
}

void QHTML::load()
{
	QString s(QFileDialog::getOpenFileName(homedir, "*", this));
	if (!s.isNull()) {
		vscroll->setValue(0); 
		analyzer->LoadFile((char*)s.ascii());
		analyzer->PackLinks();
		DrawList();
	}
}

void QHTML::contents()
{
	char ss[255];
	sprintf(ss, "%s/index.html", homedir);
	analyzer->LoadFile(ss);
	analyzer->PackLinks();
	DrawList(); 
}

void QHTML::user_guide()
{
	char ss[255];
	sprintf(ss, "%s/userg.html", homedir);
	analyzer->LoadFile(ss);
	analyzer->PackLinks();
	DrawList(); 
}

void QHTML::lang_guide()
{
	char ss[255];
	sprintf(ss, "%s/langg.html", homedir);
	analyzer->LoadFile(ss);
	analyzer->PackLinks();
	DrawList();
}

void QHTML::paintEvent(QPaintEvent *ev)
{
	if (map != NULL) {
		bitBlt(this, 0, bar->height(), map, 0, oy,
						width() - 16, height() - 16);
	}
}

void QHTML::resizeEvent(QResizeEvent *ev)
{
	DrawList();
}

void QHTML::mousePressEvent(QMouseEvent *ev)
{
	Tag *pom, *pom1;
	QString poms;
	char ss[255];

	pom = analyzer->CheckTag(ev->x(), ev->y() + oy);
	if (pom != NULL) {
		poms.sprintf(pom->tag_link);
		if (poms[0] == '"') 
			poms = poms.right(poms.length() - 1);
		if (poms[poms.length() - 1] == '"')
			poms=poms.left(poms.length() - 1);
		if (poms[0] == '#') {
			poms = poms.right(poms.length() - 1);
			pom1 = analyzer->FindAnchor((char*)poms.ascii());
			if (pom1 != NULL) {
				vscroll->setValue(pom1->y);
			}
		} else {
			sprintf(ss, "%s/%s", homedir, poms.toStdString().c_str());
			analyzer->LoadFile(ss);
			analyzer->PackLinks();
			DrawList();
		}
	}
}

/**
 * @attention Currently not in use
 */
void QHTML::back()
{
}

void QHTML::DrawList()
{
	Tag *pom;
	QPainter p;

	if (!analyzer->tags.isEmpty()) {
		cx = 5;
		cy = 15;
		map->fill(backgroundColor());
		p.begin(map);

		QListIterator<Tag *> tagsIterator(tags);

		while (tagsIterator.hasNext()) {
			pom = tagsIterator.next();
			switch(pom->tag_type) {
			case TAG_TEXT:
				p.setFont(*actual_font);
				if (cx + p.fontMetrics().width(pom->tag_text) > 
								width() - 16) {
					cx = 5;
					cy = cy + p.fontMetrics().height();
				}
				p.drawText(cx, cy, pom->tag_text);
				cx = cx + p.fontMetrics().width(pom->tag_text);
				break;
			case TAG_BREAK:
				p.setFont(*actual_font);
				cy = cy + p.fontMetrics().height();
				cx = 5;
				break;
			case TAG_ITALIC_ON:
				if (actual_font == bold)
					actual_font = bold_italic;
				else 
					actual_font = italic;

				break;
			case TAG_ITALIC_OFF:
				if (actual_font == bold_italic)
					actual_font = bold;
				else
					actual_font = normal;

				break;
			case TAG_BOLD_ON:
				if (actual_font == italic)
					actual_font = bold_italic;
				else 
					actual_font = bold;
					
				break;
			case TAG_BOLD_OFF:
				if (actual_font == bold_italic)
					actual_font = italic;
				else 
					actual_font = normal;
					
				break;
			case TAG_LINK:
				p.setFont(*actual_font);
				if (cx + p.fontMetrics().width(pom->tag_text) >
								width() - 16) {
					cx = 5;
					cy = cy + p.fontMetrics().height();
				}
				p.setPen(QColor(255, 0, 0));
				p.drawText(cx, cy, pom->tag_text);
				pom->x = cx;
				pom->y = cy + p.fontMetrics().height();
				pom->w = p.fontMetrics().width(pom->tag_text);
				pom->h = p.fontMetrics().height();
				p.setPen(QColor(0, 0, 0));
				cx = cx + p.fontMetrics().width(pom->tag_text);
				break;
			case TAG_ANCHOR:
				p.setFont(*actual_font);
				if (cx + p.fontMetrics().width(pom->tag_text) >
								width() - 16) {
					cx = 5;
					cy = cy + p.fontMetrics().height();
				}
				//p.setPen(QColor(0,255,0));
				p.drawText(cx, cy, pom->tag_text);
				pom->x = cx;
				pom->y = cy - p.fontMetrics().height();
				pom->w = p.fontMetrics().width(pom->tag_text);
				pom->h = p.fontMetrics().height();
				p.setPen(QColor(0, 0, 0));
				cx=cx + p.fontMetrics().width(pom->tag_text);
				break;
			case TAG_LIST_OFF:
			case TAG_LIST_ON:
				p.setFont(*actual_font);
				cx = 5;
				cy = cy + p.fontMetrics().height();
				break;
			case TAG_LIST_ITEM:p.setFont(*actual_font);
				p.setBrush(QBrush(QColor(0, 0, 255)));
				cx = 5;
				cy = cy + p.fontMetrics().height();
				p.drawPie(cx, cy - 5, 5, 5, 0, 5760); 
				cx = cx + 15;
				break;
			case TAG_LINE:
				p.setFont(*actual_font);
				cx = 5;
				cy = cy + p.fontMetrics().height();
				p.drawLine(cx, cy - 
					(int)(p.fontMetrics().height() / 2),
					width() - 16 - 5, cy -
					(int)(p.fontMetrics().height() / 2));
				cx = 5;
				cy = cy + p.fontMetrics().height();
				break;
			}
		}
		p.end();
		repaint();
	}
}

#include "help.moc"

int main( int argc, char **argv )
{
	QString ps;

	app = new QApplication(argc, argv);
	if (argc == 2)
		ps.sprintf(argv[1]);
	else
		ps.sprintf(".");
	QHTML cfg((char*)ps.ascii());
	//app->setStyle(WindowsStyle);
// 	app->setMainWidget(&cfg);
	cfg.show();
	return app->exec();
}
