#include "comm.h"

/** maximum special value of mark */
#define MAXMARKER       MAXINTEGER

/** maximum appetite (easily extensible ?) */
#define MAXAPPT         MAXINTEGER

/** maximum number of trace messages in line */
#define MAXTRACNT       13

/** maximum length of formal procedure header */
#define MAXHDLEN        40

/** maximum number of a system signal */
#define MAXSYSSN        62

/** maximum number of params to standard proc */
#define MAXPARAM        10

/** size of I/O transfer block in bytes */
#define IOBLOCK         0x4000

/* Object structure : */

/* Offsets from the beginning : */
/** prototype number */
#define PROTNUM         1
/** link to same size list (killed only) */
#define SHORTLINK       1
/** link to other size list (killed only) */
#define LONGLINK        2

/* Files : */
/*      appetite        0       */
/*      prot number     1       always FILEOBJECT */
/* file status */
#define FSTAT           2
/* flag to tell if file is temporary */
#define FTEMP           3
/* file type */
#define FTYPE           4
/* file name pointer */
#define FNAME           5
/** file handle */
#define FFILE           (FNAME + sizeof(char *) / sizeof(word))
/** appetite of file object */
#define APFILE          (FFILE + sizeof(FILE *) / sizeof(word))

/**
 * \defgroup ObjectOffests Offsets from the first address after object
 * @{
 */
/* static link */
#define SL              -2

/** dynamic link */
#define DL              -4

/** local control */
#define LSC             -5

/** number of times the object occurs in SL */
#define STATSL          -6

/** signal number (handlers only) */
#define SIGNR           -7

/** remote dynamic link (procedures only) */
#define RPCDL           -8

/** coroutine link (coroutine only) */
#define CL              -8

/** coroutine head (process only) */
#define CHD             -10

/** virtual scratch  (process only) */
#define VIRTSC          -12
/** @} */

/**
 * Virtual address (also formal type)
 */
typedef struct {
	/**
	 * address of dictionary item
	 * node and process index for processes
	 * or for formal types - number of arrayof
	 */
	word addr;
	/**
	 * address mark
	 * negative for processes or for formal types - actual type
	 */
	word mark;
} virtaddr;

#define loadvirt(v, a)  { word ta;              \
			  ta = (a);             \
			  (v).addr = M[ ta++ ]; \
			  (v).mark = M[ ta ]; }
#define storevirt(v, a) { word ta;              \
			  ta = (a);             \
			  M[ ta++ ] = (v).addr; \
			  M[ ta ] = (v).mark; }
			
#define MF(a)           (*( (FILE **) (M+(a)) ))
#define MN(a)           (*( (char **) (M+(a)) ))
#define MR(a)            *( (real *) (M+(a)) )


#ifdef max
#undef max
#endif

#ifdef min
#undef min
#endif

#define min(x, y)       ((x) < (y) ? (x) : (y))
#define max(x, y)       ((x) > (y) ? (x) : (y))
#define absolute(x)     ((x) >= 0 ? (x) : -(x))

/**
 * \defgroup LoglanBooleans LOGLAN's booleans
 * @{
 */
#define LFALSE          ((word)0)
#define LTRUE           (~LFALSE)
#define lbool(b)        ( (b) ? LTRUE : LFALSE )
/** @} */

/**
 * \defgroup FileTypes Type of files
 * @{
 */
/** text file */
#define TEXTF 1

/** file of char */
#define CHARF 2

/** file of integer */
#define INTF 3

/** file of real */
#define REALF 4

/** direct access file */
#define DIRECT 5
/** @} */

/**
 * \defgroup FileStatus File status
 * @{
 */
/** sequential file opened for read */
#define READING         0

/** sequential file opened for write */
#define WRITING         1

/** direct access file */
#define UPDATING        2

/** file not opened */
#define UNKNOWN         3

/** @} */

/**
 * \defgroup RuntimeErrorTypes Run time error types :
 * @{
 */
/** SL chain cut off */
#define RTESLCOF  0

/** unimplemented standard procedure */
#define RTEUNSTP  1

/** illegal attach */
#define RTEILLAT  2

/** illegal detach */
#define RTEILLDT  3

/** coroutine terminated */
#define RTECORTM  4

/** coroutine active */
#define RTECORAC  5

/** array index error */
#define RTEINVIN  6

/** incorrect array bounds */
#define RTEILLAB  7

/** improper QUA */
#define RTEINCQA  8

/** illegal assignment */
#define RTEINCAS  9

/** formal type missing */
#define RTEFTPMS 10

/** illegal kill */
#define RTEILLKL 11

/** illegal copy */
#define RTEILLCP 12

/** incompatible headers */
#define RTEINCHS 13

/** handler not found */
#define RTEHNDNF 14

/** memory overflow */
#define RTEMEMOV 15

/** formal header too long */
#define RTEFHTLG 16

/** illegal return */
#define RTEILLRT 17

/** reference to NONE */
#define RTEREFTN 18

/** division by zero */
#define RTEDIVBZ 19

/** system error */
#define RTESYSER 20

/** illegal I/O operation */
#define RTEILLIO 21

/** I/O error */
#define RTEIOERR 22

/** Cannot open file */
#define RTECNTOP 23

/** Input data format bad */
#define RTEBADFM 24

/** illegal resume */
#define RTEILLRS 25

/** too many processes on one machine */
#define RTETMPRC 26

/** invalid node number */
#define RTEINVND 27

/** negative step value */
#define RTENEGST 28

/** only process may be global */
#define RTENONGL 29
/** @} */

union value {
	unsigned int xint;
	word xword;
	real xreal;
	virtaddr xvirt;
	word xbool;
};

#define MAXINSTANCE 255

/**
 * \defgroup Variables Variables
 * @{
 */
/** main memory for code and data */
extern memory M;

/** pointer to standard proc. param list */
extern union value *param;

/** offset conversion table for compact. */
extern int offset[];

/** signal to number conversion table */
extern int scot[];

/** appetites of primitive types */
extern int primapet[];

/** instruction counter */
extern word ic;

/** previous ic for redecoding after comp. */
extern word lastic;

/** opcode of L-code instruction */
extern int opcode;

/** arguments of L-code instruction */
extern word a1, a2, a3;

/** @}*/


/**
 * \defgroup KernelVariables kernel variables for the running system
 * @{
 */
/** size of memory array for code and data */
extern word memorysize;

/** DISPLAY offset in process object */
extern word dispoff;

/** indirect DISPLAY offset in process object */
extern word disp2off;

/** DISPLAY address - physical addresses */
extern word display;

/** DISPLAY address - indirect addresses */
extern word display2;

/** pointers to current object */
extern word c1, c2;

/** main block object */
extern word mainprog;

/** offset of variable main */
extern word mnoff;

/* TODO: Does this groups ends here ?*/
/** @} */

/** TRUE if compactification message printed */
extern bool infmode;

/** TRUE if trace is printed */
extern bool debug;

/** output file for trace */
extern FILE *tracefile;

/** is graphics active ? */
bool graphics;

/** for continue execution */
extern jmp_buf contenv;

extern int internal_sock,graph_sock,net_sock;
extern int connected;
extern int GraphRes;
extern int fcol;
extern int bcol;
extern int curx;
extern int cury;
extern char ProgName[255];
extern ctx_struct my_ctx;
extern void send_to_graph(G_MESSAGE*);
extern int read_from_graph(G_MESSAGE*);
extern int read_from_net(MESSAGE*);
extern char mygname[80],mykname[80],mynname[80];
extern int DirConn[MAXINSTANCE];
