/* Loglan82 Compiler&Interpreter
     Copyright (C) 1981-1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"


#include <time.h>

/**
 * @file
 * @brief Utility routines
 */

/**
 * Compute entier (floor)
 */
word entier(double x)
{
	word i;

	if (x >= 0.0) {
		i = (word)x;
		return i;
	} else {
		i = (word)(-x);
		i = -i;
		if ((double)i <= x)
			return i;
		else
			return i - 1;
	}
}

/**
 * shift x by n bits
 */
word shift(word x, word n)
{
	if (n == 0)
		return (x);
	if (n > 0)
		return (x << n);
	else
		return ((x >> -n) & ~(~(word)0 << (8 * sizeof(word) + n)));
}


/**
 * Get ASCIIZ string from arrayof char
 */
char *asciiz(virtaddr *virt)
{
	word am;
	int len, i;
	char *cp;

	if (member(virt, &am)) {
		/* length of the string */
		len = M[am] - 3;
		/* allocate buffer for the string */
		cp = ballocate(len + 1);

		if (cp == NULL)
			errsignal(RTEMEMOV);

		for (i = 0; i < len; i++)
			cp[i] = (char) M[am + 3 + i];

		/* terminate string with 0 byte */
		cp[len] = '\0';
		return cp;
	}
	else {
		/* reference to none */
		errsignal(RTEREFTN);
	}
}

/**
 * Add extension to a file name
 */
void addext(char *fname, char *ext)
{
	char *cp;

	cp = fname;
	while (*cp != '\0' && *cp != '.')
		cp++;
	strcpy(cp, ext);
}

void usage()
{
}

/**
 * Print error message and abort
 */
void abend(char *msg)
{
	fprintf(stderr, "Error: %s\n", msg);
	exit(8);
}


/* Pseudo random number generator */

static int ranpat1 = 7, ranpat2 = 503, ranpat3 = 15661;

/**
 * Initializes random number generator
 */
void ranset()
{
	long tim;

	time(&tim);
	ranpat1 = tim % 30269;
	ranpat2 = tim % 30307;
	ranpat3 = tim % 30323;
}

/**
 * Produces next pseudo random number
 *
 * @see ranset must be invoked before running this function
 */
double prandom()
{
	int i;
	double r;

	ranpat1 = 171 * (ranpat1 % 177) - 2 * (ranpat1 / 177);
	if (ranpat1 < 0)
		ranpat1 += 30269;

	ranpat2 = 172 * (ranpat2 % 176) - 35 * (ranpat2 / 176);

	if (ranpat2 < 0)
		ranpat2 += 30307;

	ranpat3 = 170 * (ranpat3 % 178) - 63 * (ranpat3 / 178);

	if (ranpat3 < 0)
		ranpat3 += 30323;

	r = ranpat1 / 30269.0 + ranpat2 / 30307.0 + ranpat3 / 30323.0;
	i = (int)r;
	return r - i;
}

/**
 * Copy a block of memory
 *
 * @param from[in] source block to copy from
 * @param to[out] destination block to copy to
 * @param len number of character to copy
 */
void moveblock(char *from, char *to, word len)
{
	while (len-- > 0) {
		*to++ = *from++;
	}
}

/**************************************************************

#define LINE 10
void dump(word pix, word from, int len)
{
	int i;
	memory M;

	M = process[pix].M;
	while (len > 0) {
		printf("%6ld: ", (long) from);
		for (i = 0; i < LINE; i++)
			printf("%7ld", (long)M[from++]);
		putchar('\n');
		len -= LINE;
	}
}
 **************************************************************/

